"""Configuration file updater.

A configuration file consists of sections, lead by a "[section]" header,
and followed by "name: value" entries, with continuations and such in
the style of RFC 822.

The basic idea of ConfigUpdater is that a configuration file consists of
three kinds of building blocks: sections, comments and spaces for separation.
A section itself consists of three kinds of blocks: options, comments and
spaces. This gives us the corresponding data structures to describe a
configuration file.

A general block object contains the lines which were parsed and make up
the block. If a block object was not changed then during writing the same
lines that were parsed will be used to express the block. In case a block,
e.g. an option, was changed, it is marked as `updated` and its values will
be transformed into a corresponding string during an update of a
configuration file.


.. note::

   ConfigUpdater was created by starting from Python's ConfigParser source
   code and changing it according to my needs. Thus this source code
   is subject to the PSF License in a way but I am not a lawyer.
"""

import io
import os
import re
import sys
from abc import ABC
from collections import OrderedDict as _default_dict
from collections.abc import MutableMapping
from configparser import (ConfigParser, DuplicateOptionError,
                          DuplicateSectionError, Error,
                          MissingSectionHeaderError, NoOptionError,
                          NoSectionError, ParsingError)

__all__ = ["NoSectionError", "DuplicateOptionError", "DuplicateSectionError",
           "NoOptionError", "NoConfigFileReadError", "ParsingError",
           "MissingSectionHeaderError", "ConfigUpdater"]


class NoConfigFileReadError(Error):
    """Raised when no configuration file was read but update requested."""
    def __init__(self):
        super().__init__(
            "No configuration file was yet read! Use .read(...) first.")


# Used in parser getters to indicate the default behaviour when a specific
# option is not found it to raise an exception. Created to enable 'None' as
# a valid fallback value.
_UNSET = object()


class Container(ABC):
    """Abstract Mixin Class
    """
    def __init__(self, **kwargs):
        self._structure = list()
        super().__init__(**kwargs)

    @property
    def structure(self):
        return self._structure

    @property
    def last_item(self):
        if self._structure:
            return self._structure[-1]
        else:
            return None


class Block(ABC):
    """Abstract Block type holding lines

    Block objects hold original lines from the configuration file and hold
    a reference to a container wherein the object resides.
    """
    def __init__(self, container=None, **kwargs):
        self._container = container
        self.lines = []
        self._updated = False
        super().__init__(**kwargs)

    def __str__(self):
        return ''.join(self.lines)

    def __len__(self):
        return len(self.lines)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.lines == other.lines
        else:
            return False

    def add_line(self, line):
        """Add a line to the current block

        Args:
            line (str): one line to add
        """
        self.lines.append(line)
        return self

    @property
    def container(self):
        return self._container

    @property
    def add_before(self):
        """Returns a builder inserting a new block before the current block"""
        idx = self._container.structure.index(self)
        return BlockBuilder(self._container, idx)

    @property
    def add_after(self):
        """Returns a builder inserting a new block after the current block"""
        idx = self._container.structure.index(self)
        return BlockBuilder(self._container, idx+1)


class BlockBuilder(object):
    """Builder that injects blocks at a given index position."""
    def __init__(self, container, idx):
        self._container = container
        self._idx = idx

    def comment(self, text, comment_prefix='#'):
        """Creates a comment block

        Args:
            text (str): content of comment without #
            comment_prefix (str): character indicating start of comment

        Returns:
            self for chaining
        """
        comment = Comment(self._container)
        if not text.startswith(comment_prefix):
            text = "{} {}".format(comment_prefix, text)
        if not text.endswith('\n'):
            text = "{}{}".format(text, '\n')
        comment.add_line(text)
        self._container.structure.insert(self._idx, comment)
        self._idx += 1
        return self

    def section(self, section):
        """Creates a section block

        Args:
            section (str or :class:`Section`): name of section or object

        Returns:
            self for chaining
        """
        if not isinstance(self._container, ConfigUpdater):
            raise ValueError("Sections can only be added at section level!")
        if isinstance(section, str):
            # create a new section
            section = Section(section, container=self._container)
        elif not isinstance(section, Section):
            raise ValueError("Parameter must be a string or Section type!")
        if section.name in [block.name for block in self._container
                            if isinstance(block, Section)]:
            raise DuplicateSectionError(section.name)
        self._container.structure.insert(self._idx, section)
        self._idx += 1
        return self

    def space(self, newlines=1):
        """Creates a vertical space of newlines

        Args:
            newlines (int): number of empty lines

        Returns:
            self for chaining
        """
        space = Space()
        for line in range(newlines):
            space.add_line('\n')
        self._container.structure.insert(self._idx, space)
        self._idx += 1
        return self

    def option(self, key, value=None, **kwargs):
        """Creates a new option inside a section

        Args:
            key (str): key of the option
            value (str or None): value of the option
            **kwargs: are passed to the constructor of :class:`Option`

        Returns:
            self for chaining
        """
        if not isinstance(self._container, Section):
            raise ValueError("Options can only be added inside a section!")
        option = Option(key, value, container=self._container, **kwargs)
        option.value = value
        self._container.structure.insert(self._idx, option)
        self._idx += 1
        return self


class Comment(Block):
    """Comment block"""
    def __init__(self, container=None):
        super().__init__(container=container)

    def __repr__(self):
        return '<Comment>'


class Space(Block):
    """Vertical space block of new lines"""
    def __init__(self, container=None):
        super().__init__(container=container)

    def __repr__(self):
        return '<Space>'


class Section(Block, Container, MutableMapping):
    """Section block holding options

    Attributes:
        name (str): name of the section
        updated (bool): indicates name change or a new section
    """
    def __init__(self, name, container, **kwargs):
        self._name = name
        self._structure = list()
        self._updated = False
        super().__init__(container=container, **kwargs)

    def add_option(self, entry):
        """Add an Option object to the section

        Used during initial parsing mainly

        Args:
            entry (Option): key value pair as Option object
        """
        self._structure.append(entry)
        return self

    def add_comment(self, line):
        """Add a Comment object to the section

        Used during initial parsing mainly

        Args:
            line (str): one line in the comment
        """
        if not isinstance(self.last_item, Comment):
            comment = Comment(self._structure)
            self._structure.append(comment)
        self.last_item.add_line(line)
        return self

    def add_space(self, line):
        """Add a Space object to the section

        Used during initial parsing mainly

        Args:
            line (str): one line that defines the space, maybe whitespaces
        """
        if not isinstance(self.last_item, Space):
            space = Space(self._structure)
            self._structure.append(space)
        self.last_item.add_line(line)
        return self

    def _get_option_idx(self, key):
        idx = [i for i, entry in enumerate(self._structure)
               if isinstance(entry, Option) and entry.key == key]
        if idx:
            return idx[0]
        else:
            raise ValueError

    def __str__(self):
        if not self.updated:
            s = super().__str__()
        else:
            s = "[{}]\n".format(self._name)
        for entry in self._structure:
            s += str(entry)
        return s

    def __repr__(self):
        return '<Section: {}>'.format(self.name)

    def __getitem__(self, key):
        if key not in self.options():
            raise KeyError(key)
        return self._structure[self._get_option_idx(key=key)]

    def __setitem__(self, key, value):
        if key in self:
            option = self.__getitem__(key)
            option.value = value
        else:
            option = Option(key, value, container=self)
            option.value = value
            self._structure.append(option)

    def __delitem__(self, key):
        if key not in self.options():
            raise KeyError(key)
        idx = self._get_option_idx(key=key)
        del self._structure[idx]

    def __contains__(self, key):
        return key in self.options()

    def __len__(self):
        return len(self._structure)

    def __iter__(self):
        """Return all entries, not just options"""
        return self._structure.__iter__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.name == other.name and
                    self._structure == other._structure)
        else:
            return False

    def option_blocks(self):
        """Returns option blocks

        Returns:
            list: list of :class:`Option` blocks
        """
        return [entry for entry in self._structure
                if isinstance(entry, Option)]

    def options(self):
        """Returns option names

        Returns:
            list: list of option names as strings
        """
        return [option.key for option in self.option_blocks()]

    def to_dict(self):
        """Transform to dictionary

        Returns:
            dict: dictionary with same content
        """
        return {key: self.__getitem__(key).value for key in self.options()}

    @property
    def updated(self):
        """Returns if the option was changed/updated"""
        # if no lines were added, treat it as updated since we added it
        return self._updated or not self.lines

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = str(value)
        self._updated = True

    def set(self, option, value=None):
        """Set an option for chaining.

        Args:
            option (str): option name
            value (str): value, default None
        """
        option = self._container.optionxform(option)
        if option in self.options():
            self.__getitem__(option).value = value
        else:
            self.__setitem__(option, value)
        return self

    def insert_at(self, idx):
        """Returns a builder inserting a new block at the given index

        Args:
            idx (int): index where to insert
        """
        return BlockBuilder(self, idx)


class Option(Block):
    """Option block holding a key/value pair.

    Attributes:
        key (str): name of the key
        value (str): stored value
        updated (bool): indicates name change or a new section
    """
    def __init__(self, key, value, container, delimiter='=',
                 space_around_delimiters=True, line=None):
        super().__init__(container=container)
        self._key = key
        self._values = [value]
        self._value_is_none = value is None
        self._delimiter = delimiter
        self._value = None  # will be filled after join_multiline_value
        self._updated = False
        self._multiline_value_joined = False
        self._space_around_delimiters = space_around_delimiters
        if line:
            self.lines.append(line)

    def add_line(self, line):
        super().add_line(line)
        self._values.append(line.strip())

    def _join_multiline_value(self):
        if not self._multiline_value_joined and not self._value_is_none:
            # do what `_join_multiline_value` in ConfigParser would do
            self._value = '\n'.join(self._values).rstrip()
            self._multiline_value_joined = True

    def __str__(self):
        if not self.updated:
            return super().__str__()
        if self._value is None:
            return "{}{}".format(self._key, '\n')
        if self._space_around_delimiters:
            # no space is needed if we use multi-line arguments
            suffix = '' if str(self._value).startswith('\n') else ' '
            delim = " {}{}".format(self._delimiter, suffix)
        else:
            delim = self._delimiter
        return "{}{}{}{}".format(self._key, delim, self._value, '\n')

    def __repr__(self):
        return '<Option: {} = {}>'.format(self.key, self.value)

    @property
    def updated(self):
        """Returns if the option was changed/updated"""
        # if no lines were added, treat it as updated since we added it
        return self._updated or not self.lines

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._join_multiline_value()
        self._key = value
        self._updated = True

    @property
    def value(self):
        self._join_multiline_value()
        return self._value

    @value.setter
    def value(self, value):
        self._updated = True
        self._multiline_value_joined = True
        self._value = value
        self._values = [value]

    def set_values(self, values, separator='\n', indent=4*' '):
        """Sets the value to a given list of options, e.g. multi-line values

        Args:
            values (list): list of values
            separator (str): separator for values, default: line separator
            indent (str): indentation depth in case of line separator
        """
        self._updated = True
        self._multiline_value_joined = True
        self._values = values
        if separator == '\n':
            values.insert(0, '')
            separator = separator + indent
        self._value = separator.join(values)


class ConfigUpdater(Container, MutableMapping):
    """Parser for updating configuration files.

    ConfigUpdater follows the API of ConfigParser with some differences:
      * inline comments are treated as part of a key's value,
      * only a single config file can be updated at a time,
      * empty lines in values are not valid,
      * the original case of sections and keys are kept,
      * control over the position of a new section/key.

    Following features are **deliberately not** implemented:

      * interpolation of values,
      * propagation of parameters from the default section,
      * conversions of values,
      * passing key/value-pairs with ``default`` argument,
      * non-strict mode allowing duplicate sections and keys.
    """
    # Regular expressions for parsing section headers and options
    _SECT_TMPL = r"""
        \[                                 # [
        (?P<header>[^]]+)                  # very permissive!
        \]                                 # ]
        """
    _OPT_TMPL = r"""
        (?P<option>.*?)                    # very permissive!
        \s*(?P<vi>{delim})\s*              # any number of space/tab,
                                           # followed by any of the
                                           # allowed delimiters,
                                           # followed by any space/tab
        (?P<value>.*)$                     # everything up to eol
        """
    _OPT_NV_TMPL = r"""
        (?P<option>.*?)                    # very permissive!
        \s*(?:                             # any number of space/tab,
        (?P<vi>{delim})\s*                 # optionally followed by
                                           # any of the allowed
                                           # delimiters, followed by any
                                           # space/tab
        (?P<value>.*))?$                   # everything up to eol
        """
    # Compiled regular expression for matching sections
    SECTCRE = re.compile(_SECT_TMPL, re.VERBOSE)
    # Compiled regular expression for matching options with typical separators
    OPTCRE = re.compile(_OPT_TMPL.format(delim="=|:"), re.VERBOSE)
    # Compiled regular expression for matching options with optional values
    # delimited using typical separators
    OPTCRE_NV = re.compile(_OPT_NV_TMPL.format(delim="=|:"), re.VERBOSE)
    # Compiled regular expression for matching leading whitespace in a line
    NONSPACECRE = re.compile(r"\S")

    def __init__(self, allow_no_value=False, *, delimiters=('=', ':'),
                 comment_prefixes=('#', ';'), inline_comment_prefixes=None,
                 strict=True, space_around_delimiters=True):
        """Constructor of ConfigUpdater

        Args:
            allow_no_value (bool): allow keys without a value, default False
            delimiters (tuple): delimiters for key/value pairs, default =, :
            comment_prefixes (tuple): prefix of comments, default # and ;
            inline_comment_prefixes (tuple): prefix of inline comment,
                default None
            strict (bool): each section must be unique as well as every key
                within a section, default True
            space_around_delimiters (bool): add a space before and after the
                delimiter, default True
        """
        self._filename = None
        self._space_around_delimiters = space_around_delimiters

        self._dict = _default_dict  # no reason to let the user change this
        # keeping _sections to keep code aligned with ConfigParser but
        # _structure takes the actual role instead. Only use self._structure!
        self._sections = self._dict()
        self._structure = []
        self._delimiters = tuple(delimiters)
        if delimiters == ('=', ':'):
            self._optcre = self.OPTCRE_NV if allow_no_value else self.OPTCRE
        else:
            d = "|".join(re.escape(d) for d in delimiters)
            if allow_no_value:
                self._optcre = re.compile(self._OPT_NV_TMPL.format(delim=d),
                                          re.VERBOSE)
            else:
                self._optcre = re.compile(self._OPT_TMPL.format(delim=d),
                                          re.VERBOSE)
        self._comment_prefixes = tuple(comment_prefixes or ())
        self._inline_comment_prefixes = tuple(inline_comment_prefixes or ())
        self._strict = strict
        self._allow_no_value = allow_no_value
        # Options from ConfigParser that we need to set constantly
        self._empty_lines_in_values = False
        super().__init__()

    def _get_section_idx(self, name):
        idx = [i for i, entry in enumerate(self._structure)
               if isinstance(entry, Section) and entry.name == name]
        if idx:
            return idx[0]
        else:
            raise ValueError

    def read(self, filename, encoding=None):
        """Read and parse a filename.

        Args:
            filename (str): path to file
            encoding (str): encoding of file, default None
        """
        with open(filename, encoding=encoding) as fp:
            self._read(fp, filename)
        self._filename = os.path.abspath(filename)

    def read_file(self, f, source=None):
        """Like read() but the argument must be a file-like object.

        The ``f`` argument must be iterable, returning one line at a time.
        Optional second argument is the ``source`` specifying the name of the
        file being read. If not given, it is taken from f.name. If ``f`` has no
        ``name`` attribute, ``<???>`` is used.

        Args:
            f: file like object
            source (str): reference name for file object, default None
        """
        if isinstance(f, str):
            raise RuntimeError("f must be a file-like object, not string!")
        if source is None:
            try:
                source = f.name
            except AttributeError:
                source = '<???>'
        self._read(f, source)

    def read_string(self, string, source='<string>'):
        """Read configuration from a given string.

        Args:
            string (str): string containing a configuration
            source (str): reference name for file object, default '<string>'
        """
        sfile = io.StringIO(string)
        self.read_file(sfile, source)

    def optionxform(self, optionstr):
        """Converts an option key to lower case for unification

        Args:
             optionstr (str): key name

        Returns:
            str: unified option name
        """
        return optionstr.lower()

    def _update_curr_block(self, block_type):
        if not isinstance(self.last_item, block_type):
            new_block = block_type(container=self)
            self._structure.append(new_block)

    def _add_comment(self, line):
        if isinstance(self.last_item, Section):
            self.last_item.add_comment(line)
        else:
            self._update_curr_block(Comment)
            self.last_item.add_line(line)

    def _add_section(self, sectname, line):
        new_section = Section(sectname, container=self)
        new_section.add_line(line)
        self._structure.append(new_section)

    def _add_option(self, key, vi, value, line):
        entry = Option(
            key, value,
            delimiter=vi,
            container=self.last_item,
            space_around_delimiters=self._space_around_delimiters,
            line=line)
        self.last_item.add_option(entry)

    def _add_space(self, line):
        if isinstance(self.last_item, Section):
            self.last_item.add_space(line)
        else:
            self._update_curr_block(Space)
            self.last_item.add_line(line)

    def _read(self, fp, fpname):
        """Parse a sectioned configuration file.

        Each section in a configuration file contains a header, indicated by
        a name in square brackets (`[]`), plus key/value options, indicated by
        `name` and `value` delimited with a specific substring (`=` or `:` by
        default).

        Values can span multiple lines, as long as they are indented deeper
        than the first line of the value. Depending on the parser's mode, blank
        lines may be treated as parts of multiline values or ignored.

        Configuration files may include comments, prefixed by specific
        characters (`#` and `;` by default). Comments may appear on their own
        in an otherwise empty line or may be entered in lines holding values or
        section names.

        Note: This method was borrowed from ConfigParser and we keep this
        mess here as close as possible to the original messod (pardon
        this german pun) for consistency reasons and later upgrades.
        """
        self._structure = []
        elements_added = set()
        cursect = None                        # None, or a dictionary
        sectname = None
        optname = None
        lineno = 0
        indent_level = 0
        e = None                              # None, or an exception
        for lineno, line in enumerate(fp, start=1):
            comment_start = sys.maxsize
            # strip inline comments
            inline_prefixes = {p: -1 for p in self._inline_comment_prefixes}
            while comment_start == sys.maxsize and inline_prefixes:
                next_prefixes = {}
                for prefix, index in inline_prefixes.items():
                    index = line.find(prefix, index+1)
                    if index == -1:
                        continue
                    next_prefixes[prefix] = index
                    if index == 0 or (index > 0 and line[index-1].isspace()):
                        comment_start = min(comment_start, index)
                inline_prefixes = next_prefixes
            # strip full line comments
            for prefix in self._comment_prefixes:
                if line.strip().startswith(prefix):
                    comment_start = 0
                    self._add_comment(line)  # HOOK
                    break
            if comment_start == sys.maxsize:
                comment_start = None
            value = line[:comment_start].strip()
            if not value:
                if self._empty_lines_in_values:
                    # add empty line to the value, but only if there was no
                    # comment on the line
                    if (comment_start is None and
                            cursect is not None and
                            optname and
                            cursect[optname] is not None):
                        cursect[optname].append('')  # newlines added at join
                        self.last_item.last_item.add_line(line)  # HOOK
                else:
                    # empty line marks end of value
                    indent_level = sys.maxsize
                if comment_start is None:
                    self._add_space(line)
                continue
            # continuation line?
            first_nonspace = self.NONSPACECRE.search(line)
            cur_indent_level = first_nonspace.start() if first_nonspace else 0
            if (cursect is not None and optname and
                    cur_indent_level > indent_level):
                cursect[optname].append(value)
                self.last_item.last_item.add_line(line)  # HOOK
            # a section header or option header?
            else:
                indent_level = cur_indent_level
                # is it a section header?
                mo = self.SECTCRE.match(value)
                if mo:
                    sectname = mo.group('header')
                    if sectname in self._sections:
                        if self._strict and sectname in elements_added:
                            raise DuplicateSectionError(sectname, fpname,
                                                        lineno)
                        cursect = self._sections[sectname]
                        elements_added.add(sectname)
                    else:
                        cursect = self._dict()
                        self._sections[sectname] = cursect
                        elements_added.add(sectname)
                    # So sections can't start with a continuation line
                    optname = None
                    self._add_section(sectname, line)  # HOOK
                # no section header in the file?
                elif cursect is None:
                    raise MissingSectionHeaderError(fpname, lineno, line)
                # an option line?
                else:
                    mo = self._optcre.match(value)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        if not optname:
                            e = self._handle_error(e, fpname, lineno, line)
                        optname = self.optionxform(optname.rstrip())
                        if (self._strict and
                                (sectname, optname) in elements_added):
                            raise DuplicateOptionError(sectname, optname,
                                                       fpname, lineno)
                        elements_added.add((sectname, optname))
                        # This check is fine because the OPTCRE cannot
                        # match if it would set optval to None
                        if optval is not None:
                            optval = optval.strip()
                            cursect[optname] = [optval]
                        else:
                            # valueless option handling
                            cursect[optname] = None
                        self._add_option(optname, vi, optval, line)  # HOOK
                    else:
                        # a non-fatal parsing error occurred. set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        e = self._handle_error(e, fpname, lineno, line)
        # if any parsing errors occurred, raise an exception
        if e:
            raise e

    def _handle_error(self, exc, fpname, lineno, line):
        if not exc:
            exc = ParsingError(fpname)
        exc.append(lineno, repr(line))
        return exc

    def write(self, fp):
        """Write an .ini-format representation of the configuration state.

        Args:
            fp (file-like object): open file handle
        """
        fp.write(str(self))

    def update_file(self):
        """Update the read-in configuration file.
        """
        if self._filename is None:
            raise NoConfigFileReadError()
        with open(self._filename, 'w') as fb:
            self.write(fb)

    def validate_format(self, **kwargs):
        """Call ConfigParser to validate config

        Args:
            kwargs: are passed to :class:`configparser.ConfigParser`
        """
        args = dict(
            dict_type=self._dict,
            allow_no_value=self._allow_no_value,
            inline_comment_prefixes=self._inline_comment_prefixes,
            strict=self._strict,
            empty_lines_in_values=self._empty_lines_in_values
        )
        args.update(kwargs)
        parser = ConfigParser(**args)
        updated_cfg = str(self)
        parser.read_string(updated_cfg)

    def sections_blocks(self):
        """Returns all section blocks

        Returns:
            list: list of :class:`Section` blocks
        """
        return [block for block in self._structure
                if isinstance(block, Section)]

    def sections(self):
        """Return a list of section names

        Returns:
            list: list of section names
        """
        return [section.name for section in self.sections_blocks()]

    def __str__(self):
        return ''.join(str(block) for block in self._structure)

    def __getitem__(self, key):
        for section in self.sections_blocks():
            if section.name == key:
                return section
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        if not isinstance(value, Section):
            raise ValueError("Value must be of type Section!")
        if isinstance(key, str) and key in self:
            idx = self._get_section_idx(key)
            del self._structure[idx]
            self._structure.insert(idx, value)
        else:
            # name the section by the key
            value.name = key
            self.add_section(value)

    def __delitem__(self, section):
        if not self.has_section(section):
            raise KeyError(section)
        self.remove_section(section)

    def __contains__(self, key):
        return self.has_section(key)

    def __len__(self):
        """Number of all blocks, not just sections"""
        return len(self._structure)

    def __iter__(self):
        """Iterate over all blocks, not just sections"""
        return self._structure.__iter__()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._structure == other._structure
        else:
            return False

    def add_section(self, section):
        """Create a new section in the configuration.

        Raise DuplicateSectionError if a section by the specified name
        already exists. Raise ValueError if name is DEFAULT.

        Args:
            section (str or :class:`Section`): name or Section type
        """
        if section in self.sections():
            raise DuplicateSectionError(section)
        if isinstance(section, str):
            # create a new section
            section = Section(section, container=self)
        elif not isinstance(section, Section):
            raise ValueError("Parameter must be a string or Section type!")
        self._structure.append(section)

    def has_section(self, section):
        """Returns whether the given section exists.

        Args:
            section (str): name of section

        Returns:
            bool: wether the section exists
        """
        return section in self.sections()

    def options(self, section):
        """Returns list of configuration options for the named section.

        Args:
            section (str): name of section

        Returns:
            list: list of option names
        """
        if not self.has_section(section):
            raise NoSectionError(section) from None
        return self.__getitem__(section).options()

    def get(self, section, option):
        """Gets an option value for a given section.

        Args:
            section (str): section name
            option (str): option name

        Returns:
            :class:`Option`: Option object holding key/value pair
        """
        if not self.has_section(section):
            raise NoSectionError(section) from None

        section = self.__getitem__(section)
        option = self.optionxform(option)
        try:
            value = section[option]
        except KeyError:
            raise NoOptionError(option, section)

        return value

    def items(self, section=_UNSET):
        """Return a list of (name, value) tuples for options or sections.

        If section is given, return a list of tuples with (name, value) for
        each option in the section. Otherwise, return a list of tuples with
        (section_name, section_type) for each section.

        Args:
            section (str): optional section name, default UNSET

        Returns:
            list: list of :class:`Section` or :class:`Option` objects
        """
        if section is _UNSET:
            return [(sect.name, sect) for sect in self.sections_blocks()]

        section = self.__getitem__(section)
        return [(opt.key, opt) for opt in section.option_blocks()]

    def has_option(self, section, option):
        """Checks for the existence of a given option in a given section.

        Args:
            section (str): name of section
            option (str): name of option

        Returns:
            bool: whether the option exists in the given section
        """
        if section not in self.sections():
            return False
        else:
            option = self.optionxform(option)
            return option in self[section]

    def set(self, section, option, value=None):
        """Set an option.

        Args:
            section (str): section name
            option (str): option name
            value (str): value, default None
        """
        try:
            section = self.__getitem__(section)
        except KeyError:
            raise NoSectionError(section) from None
        option = self.optionxform(option)
        if option in section:
            section[option].value = value
        else:
            section[option] = value
        return self

    def remove_option(self, section, option):
        """Remove an option.

        Args:
            section (str): section name
            option (str): option name

        Returns:
            bool: whether the option was actually removed
        """
        try:
            section = self.__getitem__(section)
        except KeyError:
            raise NoSectionError(section) from None
        option = self.optionxform(option)
        existed = option in section.options()
        if existed:
            del section[option]
        return existed

    def remove_section(self, name):
        """Remove a file section.

        Args:
            name: name of the section

        Returns:
            bool: whether the section was actually removed
        """
        existed = self.has_section(name)
        if existed:
            idx = self._get_section_idx(name)
            del self._structure[idx]
        return existed

    def to_dict(self):
        """Transform to dictionary

        Returns:
            dict: dictionary with same content
        """
        return {sect: self.__getitem__(sect).to_dict()
                for sect in self.sections()}



# -*- coding: utf-8 -*-
"""
Extension that integrates cookiecutter templates into PyScaffold.

Warning:
    *Deprecation Notice* - In the next major release the Cookiecutter extension
    will be extracted into an independent package.
    After PyScaffold v4.0, you will need to explicitly install
    ``pyscaffoldext-cookiecutter`` in your system/virtualenv in order to be
    able to use it.
"""

import argparse

from ..api import Extension
from ..api.helpers import logger, register
from ..warnings import UpdateNotSupported


class Cookiecutter(Extension):
    """Additionally apply a Cookiecutter template"""
    mutually_exclusive = True

    def augment_cli(self, parser):
        """Add an option to parser that enables the Cookiecutter extension

        Args:
            parser (argparse.ArgumentParser): CLI parser object
        """
        parser.add_argument(
            self.flag,
            dest=self.name,
            action=create_cookiecutter_parser(self),
            metavar="TEMPLATE",
            help="additionally apply a Cookiecutter template. "
                 "Note that not all templates are suitable for PyScaffold. "
                 "Please refer to the docs for more information.")

    def activate(self, actions):
        """Register before_create hooks to generate project using Cookiecutter

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        # `get_default_options` uses passed options to compute derived ones,
        # so it is better to prepend actions that modify options.
        actions = register(actions, enforce_cookiecutter_options,
                           before='get_default_options')

        # `apply_update_rules` uses CWD information,
        # so it is better to prepend actions that modify it.
        actions = register(actions, create_cookiecutter,
                           before='apply_update_rules')

        return actions


def create_cookiecutter_parser(obj_ref):
    """Create a Cookiecutter parser.

    Args:
        obj_ref (Extension): object reference to the actual extension

    Returns:
        NamespaceParser: parser for namespace cli argument
    """
    class CookiecutterParser(argparse.Action):
        """Consumes the values provided, but also append the extension function
        to the extensions list.
        """

        def __call__(self, parser, namespace, values, option_string=None):
            # First ensure the extension function is stored inside the
            # 'extensions' attribute:
            extensions = getattr(namespace, 'extensions', [])
            extensions.append(obj_ref)
            setattr(namespace, 'extensions', extensions)

            # Now the extra parameters can be stored
            setattr(namespace, self.dest, values)

            # save the cookiecutter cli argument for later
            obj_ref.args = values

    return CookiecutterParser


def enforce_cookiecutter_options(struct, opts):
    """Make sure options reflect the cookiecutter usage.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        struct, opts: updated project representation and options
    """
    opts['force'] = True

    return struct, opts


def create_cookiecutter(struct, opts):
    """Create a cookie cutter template

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        struct, opts: updated project representation and options
    """
    if opts.get('update'):
        logger.warning(UpdateNotSupported(extension='cookiecutter'))
        return struct, opts

    try:
        from cookiecutter.main import cookiecutter
    except Exception as e:
        raise NotInstalled from e

    extra_context = dict(full_name=opts['author'],
                         author=opts['author'],
                         email=opts['email'],
                         project_name=opts['project'],
                         package_name=opts['package'],
                         repo_name=opts['package'],
                         project_short_description=opts['description'],
                         release_date=opts['release_date'],
                         version='unknown',  # will be replaced later
                         year=opts['year'])

    if 'cookiecutter' not in opts:
        raise MissingTemplate

    logger.report('run', 'cookiecutter ' + opts['cookiecutter'])
    if not opts.get('pretend'):
        cookiecutter(opts['cookiecutter'],
                     no_input=True,
                     extra_context=extra_context)

    return struct, opts


class NotInstalled(RuntimeError):
    """This extension depends on the ``cookiecutter`` package."""

    DEFAULT_MESSAGE = ("cookiecutter is not installed, "
                       "run: pip install cookiecutter")

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super(NotInstalled, self).__init__(message, *args, **kwargs)


class MissingTemplate(RuntimeError):
    """A cookiecutter template (git url) is required."""

    DEFAULT_MESSAGE = "missing `cookiecutter` option"

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super(MissingTemplate, self).__init__(message, *args, **kwargs)



# -*- coding: utf-8 -*-
"""
Templates for all files of a project's scaffold
"""

import os
import string
from pkg_resources import resource_string

from ..utils import get_setup_requires_version
from .. import __version__ as pyscaffold_version
from ..contrib.configupdater import ConfigUpdater


#: All available licences
licenses = {"affero": "license_affero_3.0",
            "apache": "license_apache",
            "artistic": "license_artistic_2.0",
            "cc0": "license_cc0_1.0",
            "eclipse": "license_eclipse_1.0",
            "gpl2": "license_gpl_2.0",
            "gpl3": "license_gpl_3.0",
            "isc": "license_isc",
            "lgpl2": "license_lgpl_2.1",
            "lgpl3": "license_lgpl_3.0",
            "mit": "license_mit",
            "mozilla": "license_mozilla",
            "new-bsd": "license_new_bsd",
            "none": "license_none",
            "proprietary": "license_none",
            "public-domain": "license_public_domain",
            "simple-bsd": "license_simplified_bsd"}


def get_template(name):
    """Retrieve the template by name

    Args:
        name: name of template

    Returns:
        :obj:`string.Template`: template
    """
    file_name = "{name}.template".format(name=name)
    data = resource_string(__name__, file_name)
    # we assure that line endings are converted to '\n' for all OS
    data = data.decode(encoding="utf-8").replace(os.linesep, '\n')
    return string.Template(data)


def setup_py(opts):
    """Template of setup.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("setup_py")
    return template.safe_substitute(opts)


def setup_cfg(opts):
    """Template of setup.cfg

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("setup_cfg")
    opts["setup_requires_str"] = get_setup_requires_version()
    cfg_str = template.substitute(opts)

    updater = ConfigUpdater()
    updater.read_string(cfg_str)

    # add `classifiers`
    (updater["metadata"]["platforms"].add_after
     .comment("Add here all kinds of additional classifiers as defined under")
     .comment("https://pypi.python.org/pypi?%3Aaction=list_classifiers")
     .option("classifiers"))
    updater["metadata"]["classifiers"].set_values(opts["classifiers"])

    # add `install_requires`
    setup_requires = updater["options"]["setup_requires"]
    if opts["requirements"]:
        setup_requires.add_after.option("install_requires")
        updater["options"]["install_requires"].set_values(opts["requirements"])
    else:
        (setup_requires.add_after
         .comment("Add here dependencies of your project "
                  "(semicolon/line-separated), e.g.")
         .comment("install_requires = numpy; scipy"))

    # fill [pyscaffold] section used for later updates
    pyscaffold = updater["pyscaffold"]
    pyscaffold["version"] = pyscaffold_version
    pyscaffold["package"] = opts["package"]
    if opts["cli_params"]["extensions"]:
        pyscaffold.set("extensions")
        pyscaffold["extensions"].set_values(opts["cli_params"]["extensions"])
        for extension, args in opts["cli_params"]["args"].items():
            pyscaffold[extension] = args

    return str(updater)


def gitignore(opts):
    """Template of .gitignore

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("gitignore")
    return template.substitute(opts)


def gitignore_empty(opts):
    """Template of empty .gitignore

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("gitignore_empty")
    return template.substitute(opts)


def sphinx_conf(opts):
    """Template of conf.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_conf")
    return template.substitute(opts)


def sphinx_index(opts):
    """Template of index.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_index")
    return template.substitute(opts)


def sphinx_license(opts):
    """Template of license.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_license")
    return template.substitute(opts)


def sphinx_authors(opts):
    """Template of authors.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_authors")
    return template.substitute(opts)


def sphinx_changelog(opts):
    """Template of changelog.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_changelog")
    return template.substitute(opts)


def sphinx_makefile(opts):
    """Template of Sphinx's Makefile

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("sphinx_makefile")
    return template.safe_substitute(opts)


def readme(opts):
    """Template of README.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("readme")
    return template.substitute(opts)


def authors(opts):
    """Template of AUTHORS.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("authors")
    return template.substitute(opts)


def requirements(opts):
    """Template of requirements.txt

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("requirements")
    reqs = "\n".join(opts["requirements"]) or "#"
    opts["requirements_str"] = reqs
    return template.substitute(opts)


def license(opts):
    """Template of LICENSE.txt

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template(licenses[opts["license"]])
    return template.substitute(opts)


def init(opts):
    """Template of __init__.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    if opts["package"] == opts["project"]:
        opts["distribution"] = "__name__"
    else:
        opts["distribution"] = "'{}'".format(opts["project"])
    template = get_template("__init__")
    return template.substitute(opts)


def coveragerc(opts):
    """Template of .coveragerc

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("coveragerc")
    return template.substitute(opts)


def tox(opts):
    """Template of tox.ini

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("tox_ini")
    return template.substitute(opts)


def travis(opts):
    """Template of .travis.yml

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("travis")
    return template.safe_substitute(opts)


def travis_install(opts):
    """Template of travis_install.sh

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("travis_install")
    return template.safe_substitute(opts)


def gitlab_ci(opts):
    """Template of .gitlab-ci.yml

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("gitlab_ci")
    return template.safe_substitute(opts)


def pre_commit_config(opts):
    """Template of .pre-commit-config.yaml

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("pre-commit-config")
    return template.safe_substitute(opts)


def isort_cfg(opts):
    """Template of .isort.cfg

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("isort_cfg")
    return template.safe_substitute(opts)


def namespace(opts):
    """Template of __init__.py defining a namespace package

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("namespace")
    return template.substitute(opts)


def skeleton(opts):
    """Template of skeleton.py defining a basic console script

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("skeleton")
    return template.substitute(opts)


def test_skeleton(opts):
    """Template of unittest for skeleton.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("test_skeleton")
    return template.substitute(opts)


def changelog(opts):
    """Template of CHANGELOG.rst

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("changelog")
    return template.substitute(opts)


def conftest_py(opts):
    """Template of conftest.py

    Args:
        opts: mapping parameters as dictionary

    Returns:
        str: file content as string
    """
    template = get_template("conftest_py")
    return template.substitute(opts)



# -*- coding: utf-8 -*-
import sys
import os
from pkg_resources import get_distribution, DistributionNotFound
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from autodiffcst.AD_vec import *


try:
    dist_name = 'autodiffCST'
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = 'unknown'
finally:
    del get_distribution, DistributionNotFound



import math
import numbers
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import numpy as np
import sympy as sp
from autodiffcst import AD
# import autodiffcst.AD as AD


def choose(n, k):
    """
    A helper function that gives the value of n choose k, according to math definition

            Parameters:
                    n, k: both natural numbers, invalid input cases handled by numpy
            
            Returns:
                    the arithmetic value of n choose k, a scalar
            
            Example:
            >>> choose(5,4)
            5.0 
    """
    return np.math.factorial(n) / (np.math.factorial(k) * np.math.factorial(n - k))


def chain_rule(ad, new_val, der, der2, higher_der=None):
    """
    Applies chain rule to returns a new AD object with correct value and derivatives.

            Parameters:
                    ad (AD): An AD object
                    new_val (float): Value of the new AD object
                    der (float): Derivative of the outer function in chain rule

            Returns:
                    new_ad (AD): a new AD object with correct value and derivatives

            Example:
            >>> import AD as AD
            >>> x = AD.AD(2,order=5)
            >>> newad = 5*x**3+2
            >>> higherde = np.array([60,60,30,0,0])
            >>> chain_rule(x, 42, 60, 60, higher_der=higherde)
            AD(value: [42], derivatives: [60.])
    """
    new_der = der * ad.der
    new_der2 = der * ad.der2 + der2 * np.matmul(np.array([ad.der]).T, np.array([ad.der]))
    if ad.higher is None:
        new_ad = AD.AD(new_val, tag=ad.tag, der=new_der, der2=new_der2)
    else:
        new_higher_der = np.array([0.0] * len(ad.higher))
        new_higher_der[0] = new_der
        new_higher_der[1] = new_der2
        for i in range(2, len(ad.higher)):
            n = i + 1
            sum = 0
            for k in range(1, n + 1):
                sum += higher_der[k - 1] * sp.bell(n, k, ad.higher[0:n - k + 1])
            new_higher_der[i] = sum
        new_ad = AD.AD(new_val, tag=ad.tag, der=new_der, der2=new_der2, order=len(ad.higher))
        new_ad.higher = new_higher_der

    return new_ad


def abs(ad):
    """
    Returns the new AD object after applying absolute value function.

            Parameters:
                    ad (AD): An AD object to be applied absolute value function on

            Returns:
                    new_ad (AD): the new AD object after applying absolute value function

            Example:
            >>> x = AD.AD(-2,order=5) 
            >>> abs(x)
            AD(value: [2], derivatives: [-1.])
    """
    if isinstance(ad, AD.AD):
        new_val = np.abs((ad.val))
        der = np.array([0] * len(ad))
        der2 = np.array([0] * len(ad))

        def get_der(v):
            if v > 0:
                return 1
            elif v < 0:
                return -1
            else:
                raise Exception("Derivative undefined")

        for i in range(0, len(ad.val)):
            if isinstance(ad.val[i], numbers.Integral):
                der[i] = get_der(ad.val[i])
            else:
                #print(type(ad.val[i]))
                sub_der = np.array([get_der(v) for v in ad.val[i]])
                der[i] = sub_der
                der2[i] = np.array([0] * len(ad.val[i]))

        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([0.0] * len(ad.higher))
            higher_der[0] = der
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.abs(ad)
        except:
            raise TypeError("Your input is not valid.")


def exp(ad):
    """
    Returns the new AD object after applying exponential function.

            Parameters:
                    ad (AD): An AD object to be applied exponential function on

            Returns:
                    new_ad (AD): the new AD object after applying exponential function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> exp(x)
            AD(value: [7.3890561], derivatives: [7.3890561])
    """
    if isinstance(ad, AD.AD):
        new_val = np.exp(ad.val)
        der = new_val
        der2 = new_val
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([new_val] * len(ad.higher))
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.exp(ad)
        except:
            raise TypeError("Your input is not valid.")


def fact_ad(x, n):
    """
    Returns x(x-1)(x-2)...(x-n+1), the product of n terms, factorial-like operation

            Parameters:
                    x, n: two scalars

            Returns:
                    x(x-1)(x-2)...(x-n+1): scalar

            Example:
            >>> fact_ad(5,4)
            120
    """
    prod = 1
    for i in range(n):
        prod = prod * (x - i)
    return prod


def log(ad):  # consider different base?
    """
    Returns the new AD object after applying log(base e) function.

            Parameters:
                    ad (AD): An AD object to be applied log(base e) function on

            Returns:
                    new_ad (AD): the new AD object after applying log(base e) function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> log(x)
            AD(value: [0.69314718], derivatives: [0.5])
    """
    if isinstance(ad, AD.AD):
        new_val = np.log(ad.val)
        der = 1 / ad.val
        der2 = -1 / ad.val ** 2
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            # starting from the first derivative: x**-1
            higher_der = np.array([0.0] * len(ad.higher))
            higher_der[0] = der
            higher_der[1] = der2
            for i in range(2, len(ad.higher)):
                n = i + 1
                coef = fact_ad(-1, n - 1)
                mainval = np.power(ad.val[0], float(-n))
                # mainval = math.pow(ad.val[0], -n)
                higher_der[i] = coef * mainval
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.log(ad)
        except:
            raise TypeError("Your input is not valid.")





def sqrt(ad):
    """
    Returns the new AD object after applying square root function.

            Parameters:
                    ad (AD): An AD object to be applied square root function on

            Returns:
                    new_ad (AD): the new AD object after applying square root function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> sqrt(x)
            AD(value: [1.41421356], derivatives: [0.35355339])
    """
    return ad ** 0.5

# trig
def sin(ad):
    """
    Returns the new AD object after applying sine function.

            Parameters:
                    ad (AD): An AD object to be applied sine function on

            Returns:
                    new_ad (AD): the new AD object after applying sine function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> sin(x)
            AD(value: [0.90929743], derivatives: [-0.41614684])
    """
    if isinstance(ad, AD.AD):
        new_val = sin(ad.val)
        der = cos(ad.val)
        der2 = -new_val
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([der, der2, -der, -der2] * int(np.ceil(len(ad.higher) / 4)))
            higher_der = higher_der[0:len(ad.higher)]
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.sin(ad)
        except:
            raise TypeError("Your input is not valid.")


def cos(ad):
    """
    Returns the new AD object after applying cosine function.

            Parameters:
                    ad (AD): An AD object to be applied cosine function on

            Returns:
                    new_ad (AD): the new AD object after applying cosine function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> cos(x)
            AD(value: [-0.41614684], derivatives: [-0.90929743])
    """
    if isinstance(ad, AD.AD):
        new_val = cos(ad.val)
        der = -sin(ad.val)
        der2 = -cos(ad.val)
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([der, der2, -der, -der2] * int(np.ceil(len(ad.higher) / 4)))
            higher_der = higher_der[0:len(ad.higher)]
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.cos(ad)
        except:
            raise TypeError("Your input is not valid.")


def tan(ad):
    """
    Returns the new AD object after applying tangent function.

            Parameters:
                    ad (AD): An AD object to be applied tangent function on

            Returns:
                    new_ad (AD): the new AD object after applying tangent function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> tan(x)
            AD(value: [-2.18503986], derivatives: [5.7743992])
    """
    if isinstance(ad, AD.AD):
        return sin(ad) / cos(ad)
    else:
        try:
            return np.tan(ad)
        except:
            raise TypeError("Your input is not valid.")


def sec(num):
    """
    Returns the new AD object after applying secant function.
            Parameters:
                    num: a number or array
            Returns:
                    the result after applying secant function

            Example:
            >>> sec(np.array([1,2,3]))
            array([ 1.85081572, -2.40299796, -1.01010867])
    """
    try:
        return 1 / np.cos(num)
    except:
        raise TypeError("sec function can only handle number or array.")


# hyperbolic trig
def sinh(ad):
    """
    Returns the new AD object after applying hyperbolic sine function.

            Parameters:
                    ad (AD): An AD object to be applied hyperbolic sine function on

            Returns:
                    new_ad (AD): the new AD object after applying hyperbolic sine function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> sinh(x)
            AD(value: [3.62686041], derivatives: [3.76219569])
    """
    if isinstance(ad, AD.AD):
        new_val = sinh(ad.val)
        der = cosh(ad.val)
        der2 = sinh(ad.val)
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([der, der2, der, der2] * int(np.ceil(len(ad.higher) / 4)))
            higher_der = higher_der[0:len(ad.higher)]
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.sinh(ad)
        except:
            raise TypeError("Your input is not valid.")


def cosh(ad):
    """
    Returns the new AD object after applying hyperbolic cosine function.

            Parameters:
                    ad (AD): An AD object to be applied hyperbolic cosine function on

            Returns:
                    new_ad (AD): the new AD object after applying hyperbolic cosine function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> cosh(x)
            AD(value: [3.76219569], derivatives: [3.62686041])
    """
    if isinstance(ad, AD.AD):
        new_val = cosh(ad.val)
        der = sinh(ad.val)
        der2 = new_val
        if ad.higher is None:
            return chain_rule(ad, new_val, der, der2)
        else:
            higher_der = np.array([der, der2, der, der2] * int(np.ceil(len(ad.higher) / 4)))
            higher_der = higher_der[0:len(ad.higher)]
            return chain_rule(ad, new_val, der, der2, higher_der)
    else:
        try:
            return np.cosh(ad)
        except:
            raise TypeError("Your input is not valid.")


def sech(ad):
    """
    Returns the new AD object after applying hyperbolic secant function.

            Parameters:
                    ad (AD): An AD object to be applied hyperbolic secant function on

            Returns:
                    new_ad (AD): the new AD object after applying hyperbolic secant function

            Example:
            >>> sech(np.array([1,2,3]))
            array([0.64805427, 0.26580223, 0.09932793])
    """
    try:
        return 1 / np.cosh(ad)
    except:
        raise TypeError("sec function can only handle number or array.")


def tanh(ad):
    """
    Returns the new AD object after applying hyperbolic tangent function.

            Parameters:
                    ad (AD): An AD object to be applied hyperbolic tangent function on

            Returns:
                    new_ad (AD): the new AD object after applying hyperbolic tangent function

            Example:
            >>> x = AD.AD(2,order=5)
            >>> tanh(x)
            AD(value: [0.96402758], derivatives: [0.07065082])
    """
    if isinstance(ad, AD.AD):
        return sinh(ad)/cosh(ad)
    else:
        try:
            return np.tanh(ad)
        except:
            raise TypeError("Your input is not valid.")



# -*- coding: utf-8 -*-
"""
Exposed API for accessing PyScaffold via Python.
"""

import os
from datetime import date, datetime
from functools import reduce

import pyscaffold

from .. import info, repo, utils
from ..exceptions import (
    DirectoryAlreadyExists,
    DirectoryDoesNotExist,
    GitDirtyWorkspace,
    InvalidIdentifier
)
from ..log import logger
from ..structure import (
    create_structure,
    define_structure
)
from ..update import (
    apply_update_rules,
    version_migration,
    invoke_action
)
from . import helpers


# -------- Extension Main Class --------

class Extension(object):
    """Base class for PyScaffold's extensions

    Args:
        name (str): How the extension should be named. Default: name of class
            By default, this value is used to create the activation flag in
            PyScaffold cli.
    """
    mutually_exclusive = False

    def __init__(self, name):
        self.name = name
        self.args = None

    @property
    def flag(self):
        return '--{flag}'.format(flag=utils.dasherize(self.name))

    def augment_cli(self, parser):
        """Augments the command-line interface parser

        A command line argument ``--FLAG`` where FLAG=``self.name`` is added
        which appends ``self.activate`` to the list of extensions. As help
        text the docstring of the extension class is used.
        In most cases this method does not need to be overwritten.

        Args:
            parser: current parser object
        """
        help = self.__doc__[0].lower() + self.__doc__[1:]

        parser.add_argument(
            self.flag,
            help=help,
            dest="extensions",
            action="append_const",
            const=self)
        return self

    def activate(self, actions):
        """Activates the extension by registering its functionality

        Args:
            actions (list): list of action to perform

        Returns:
            list: updated list of actions
        """
        raise NotImplementedError(
            "Extension {} has no actions registered".format(self.name))

    @staticmethod
    def register(*args, **kwargs):
        """Shortcut for :obj:`helpers.register`"""
        return helpers.register(*args, **kwargs)

    @staticmethod
    def unregister(*args, **kwargs):
        """Shortcut for :obj:`helpers.unregister`"""
        return helpers.unregister(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Just delegating to :obj:`self.activate`"""
        return self.activate(*args, **kwargs)


# -------- Actions --------

DEFAULT_OPTIONS = {'update': False,
                   'force': False,
                   'description': 'Add a short description here!',
                   'url': 'https://github.com/pyscaffold/pyscaffold/',
                   'license': 'mit',
                   'version': pyscaffold.__version__,
                   'classifiers': ['Development Status :: 4 - Beta',
                                   'Programming Language :: Python'],
                   }


def discover_actions(extensions):
    """Retrieve the action list.

    This is done by concatenating the default list with the one generated after
    activating the extensions.

    Args:
        extensions (list): list of functions responsible for activating the
        extensions.

    Returns:
        list: scaffold actions.
    """
    actions = DEFAULT_ACTIONS.copy()

    # Order the extensions lexicographically which is needed for determinism,
    # also internal before external "pyscaffold.*" < "pyscaffoldext.*"
    def sort_by_qual_name(ext):
        return '.'.join([ext.__module__, ext.__class__.__qualname__])

    extensions = sorted(extensions, key=sort_by_qual_name)
    # Activate the extensions
    return reduce(lambda acc, f: _activate(f, acc), extensions, actions)


def get_default_options(struct, opts):
    """Compute all the options that can be automatically derived.

    This function uses all the available information to generate sensible
    defaults. Several options that can be derived are computed when possible.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        dict, dict: project representation and options with default values set

    Raises:
        :class:`~.DirectoryDoesNotExist`: when PyScaffold is told to
            update an nonexistent directory
        :class:`~.GitNotInstalled`: when git command is not available
        :class:`~.GitNotConfigured`: when git does not know user information

    Note:
        This function uses git to determine some options, such as author name
        and email.
    """
    # This function uses information from git, so make sure it is available
    info.check_git()

    given_opts = opts
    # Initial parameters that need to be provided also during an update
    opts = DEFAULT_OPTIONS.copy()
    opts.update(given_opts)
    opts.setdefault('package', utils.make_valid_identifier(opts['project']))
    opts.setdefault('author', info.username())
    opts.setdefault('email', info.email())
    opts.setdefault('release_date', date.today().strftime('%Y-%m-%d'))
    # All kinds of derived parameters
    year = datetime.strptime(opts['release_date'], '%Y-%m-%d').year
    opts.setdefault('year', year)
    opts.setdefault('title',
                    '='*len(opts['project']) + '\n' + opts['project'] + '\n' +
                    '='*len(opts['project']))

    # Initialize empty list of all requirements and extensions
    # (since not using deep_copy for the DEFAULT_OPTIONS, better add compound
    # values inside this function)
    opts.setdefault('requirements', list())
    opts.setdefault('extensions', list())
    opts.setdefault('root_pkg', opts['package'])
    opts.setdefault('qual_pkg', opts['package'])
    opts.setdefault('cli_params', {'extensions': list(), 'args': dict()})
    opts.setdefault('pretend', False)

    return struct, opts


def verify_options_consistency(struct, opts):
    """Perform some sanity checks about the given options.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        dict, dict: updated project representation and options
    """
    if not utils.is_valid_identifier(opts['package']):
        raise InvalidIdentifier(
            "Package name {} is not a valid "
            "identifier.".format(opts['package']))

    if opts['update'] and not opts['force']:
        if not info.is_git_workspace_clean(opts['project']):
            raise GitDirtyWorkspace

    return struct, opts


def verify_project_dir(struct, opts):
    """Check if PyScaffold can materialize the project dir structure.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        dict, dict: updated project representation and options
    """
    if os.path.exists(opts['project']):
        if not opts['update'] and not opts['force']:
            raise DirectoryAlreadyExists(
                "Directory {dir} already exists! Use the `update` option to "
                "update an existing project or the `force` option to "
                "overwrite an existing directory.".format(dir=opts['project']))
    elif opts['update']:
        raise DirectoryDoesNotExist(
            "Project {project} does not exist and thus cannot be "
            "updated!".format(project=opts['project']))

    return struct, opts


def init_git(struct, opts):
    """Add revision control to the generated files.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`.
        opts (dict): given options, see :obj:`create_project` for
            an extensive list.

    Returns:
        dict, dict: updated project representation and options
    """
    if not opts['update'] and not repo.is_git_repo(opts['project']):
        repo.init_commit_repo(opts['project'], struct,
                              log=True, pretend=opts.get('pretend'))

    return struct, opts


# -------- API --------

DEFAULT_ACTIONS = [
    get_default_options,
    verify_options_consistency,
    define_structure,
    verify_project_dir,
    apply_update_rules,
    version_migration,
    create_structure,
    init_git
]


def create_project(opts=None, **kwargs):
    """Create the project's directory structure

    Args:
        opts (dict): options of the project
        **kwargs: extra options, passed as keyword arguments

    Returns:
        tuple: a tuple of `struct` and `opts` dictionary

    Valid options include:

    :Naming:                - **project** (*str*)
                            - **package** (*str*)

    :Package Information:   - **author** (*str*)
                            - **email** (*str*)
                            - **release_date** (*str*)
                            - **year** (*str*)
                            - **title** (*str*)
                            - **description** (*str*)
                            - **url** (*str*)
                            - **classifiers** (*str*)
                            - **requirements** (*list*)

    :PyScaffold Control:    - **update** (*bool*)
                            - **force** (*bool*)
                            - **pretend** (*bool*)
                            - **extensions** (*list*)

    Some of these options are equivalent to the command line options, others
    are used for creating the basic python package meta information, but the
    last tree can change the way PyScaffold behaves.

    When the **force** flag is ``True``, existing files will be overwritten.
    When the **update** flag is ``True``, PyScaffold will consider that some
    files can be updated (usually the packaging boilerplate),
    but will keep others intact.
    When the **pretend** flag is ``True``, the project will not be
    created/updated, but the expected outcome will be logged.

    Finally, the **extensions** list may contain any function that follows the
    `extension API <../extensions>`_. Note that some PyScaffold features, such
    as travis, tox and pre-commit support, are implemented as built-in
    extensions.  In order to use these features it is necessary to include the
    respective functions in the extension list.  All built-in extensions are
    accessible via :mod:`pyscaffold.extensions` submodule.

    Note that extensions may define extra options. For example, built-in
    cookiecutter extension define a ``cookiecutter`` option that
    should be the address to the git repository used as template.
    """
    opts = opts if opts else {}
    opts.update(kwargs)

    actions = discover_actions(opts.get('extensions', []))

    # call the actions to generate final struct and opts
    struct = {}
    struct, opts = reduce(lambda acc, f: invoke_action(f, *acc),
                          actions, (struct, opts))
    return struct, opts


# -------- Auxiliary functions --------

def _activate(extension, actions):
    """Activate extension with proper logging."""
    logger.report('activate', extension.__module__)
    with logger.indent():
        actions = extension(actions)

    return actions



""" configuration """
from __future__ import print_function, unicode_literals
import os
import re
import warnings

from .utils import trace

DEFAULT_TAG_REGEX = r"^(?:[\w-]+-)?(?P<version>[vV]?\d+(?:\.\d+){0,2}[^\+]+)(?:\+.*)?$"
DEFAULT_VERSION_SCHEME = "version_scheme"


def _check_tag_regex(value):
    if not value:
        value = DEFAULT_TAG_REGEX
    regex = re.compile(value)

    group_names = regex.groupindex.keys()
    if regex.groups == 0 or (regex.groups > 1 and "version" not in group_names):
        warnings.warn(
            "Expected tag_regex to contain a single match group or a group named"
            " 'version' to identify the version part of any tag."
        )

    return regex


def _check_absolute_root(root, relative_to):
    if relative_to:
        if os.path.isabs(root) and not root.startswith(relative_to):
            warnings.warn(
                "absolute root path '%s' overrides relative_to '%s'"
                % (root, relative_to)
            )
        root = os.path.join(os.path.dirname(relative_to), root)
    return os.path.abspath(root)


class Configuration(object):
    """ Global configuration model """

    _root = None
    version_scheme = None
    local_scheme = None
    write_to = None
    write_to_template = None
    fallback_version = None
    _relative_to = None
    parse = None
    _tag_regex = None
    _absolute_root = None

    def __init__(self, relative_to=None, root="."):
        # TODO:
        self._relative_to = relative_to
        self._root = "."

        self.root = root
        self.version_scheme = DEFAULT_VERSION_SCHEME
        self.local_scheme = "node-and-date"
        self.write_to = ""
        self.write_to_template = None
        self.fallback_version = None
        self.fallback_root = "."
        self.parse = None
        self.tag_regex = DEFAULT_TAG_REGEX
        self.git_describe_command = None

    @property
    def fallback_root(self):
        return self._fallback_root

    @fallback_root.setter
    def fallback_root(self, value):
        self._fallback_root = os.path.abspath(value)

    @property
    def absolute_root(self):
        return self._absolute_root

    @property
    def relative_to(self):
        return self._relative_to

    @relative_to.setter
    def relative_to(self, value):
        self._absolute_root = _check_absolute_root(self._root, value)
        self._relative_to = value
        trace("root", repr(self._absolute_root))

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, value):
        self._absolute_root = _check_absolute_root(value, self._relative_to)
        self._root = value
        trace("root", repr(self._absolute_root))

    @property
    def tag_regex(self):
        return self._tag_regex

    @tag_regex.setter
    def tag_regex(self, value):
        self._tag_regex = _check_tag_regex(value)



"""Activate coverage at python startup if appropriate.

The python site initialisation will ensure that anything we import
will be removed and not visible at the end of python startup.  However
we minimise all work by putting these init actions in this separate
module and only importing what is needed when needed.

For normal python startup when coverage should not be activated the pth
file checks a single env var and does not import or call the init fn
here.

For python startup when an ancestor process has set the env indicating
that code coverage is being collected we activate coverage based on
info passed via env vars.
"""
import atexit
import os
import signal

_active_cov = None


def multiprocessing_start(_):
    global _active_cov
    cov = init()
    if cov:
        _active_cov = cov
        multiprocessing.util.Finalize(None, cleanup, exitpriority=1000)


try:
    import multiprocessing.util
except ImportError:
    pass
else:
    multiprocessing.util.register_after_fork(multiprocessing_start, multiprocessing_start)


def init():
    # Only continue if ancestor process has set everything needed in
    # the env.
    global _active_cov

    cov_source = os.environ.get('COV_CORE_SOURCE')
    cov_config = os.environ.get('COV_CORE_CONFIG')
    cov_datafile = os.environ.get('COV_CORE_DATAFILE')
    cov_branch = True if os.environ.get('COV_CORE_BRANCH') == 'enabled' else None

    if cov_datafile:
        if _active_cov:
            cleanup()
        # Import what we need to activate coverage.
        import coverage

        # Determine all source roots.
        if cov_source in os.pathsep:
            cov_source = None
        else:
            cov_source = cov_source.split(os.pathsep)
        if cov_config == os.pathsep:
            cov_config = True

        # Activate coverage for this process.
        cov = _active_cov = coverage.Coverage(
            source=cov_source,
            branch=cov_branch,
            data_suffix=True,
            config_file=cov_config,
            auto_data=True,
            data_file=cov_datafile
        )
        cov.load()
        cov.start()
        cov._warn_no_data = False
        cov._warn_unimported_source = False
        return cov


def _cleanup(cov):
    if cov is not None:
        cov.stop()
        cov.save()
        cov._auto_save = False  # prevent autosaving from cov._atexit in case the interpreter lacks atexit.unregister
        try:
            atexit.unregister(cov._atexit)
        except Exception:
            pass


def cleanup():
    global _active_cov
    global _cleanup_in_progress
    global _pending_signal

    _cleanup_in_progress = True
    _cleanup(_active_cov)
    _active_cov = None
    _cleanup_in_progress = False
    if _pending_signal:
        pending_singal = _pending_signal
        _pending_signal = None
        _signal_cleanup_handler(*pending_singal)


multiprocessing_finish = cleanup  # in case someone dared to use this internal

_previous_handlers = {}
_pending_signal = None
_cleanup_in_progress = False


def _signal_cleanup_handler(signum, frame):
    global _pending_signal
    if _cleanup_in_progress:
        _pending_signal = signum, frame
        return
    cleanup()
    _previous_handler = _previous_handlers.get(signum)
    if _previous_handler == signal.SIG_IGN:
        return
    elif _previous_handler and _previous_handler is not _signal_cleanup_handler:
        _previous_handler(signum, frame)
    elif signum == signal.SIGTERM:
        os._exit(128 + signum)
    elif signum == signal.SIGINT:
        raise KeyboardInterrupt()


def cleanup_on_signal(signum):
    previous = signal.getsignal(signum)
    if previous is not _signal_cleanup_handler:
        _previous_handlers[signum] = previous
        signal.signal(signum, _signal_cleanup_handler)


def cleanup_on_sigterm():
    cleanup_on_signal(signal.SIGTERM)



# -*- coding: utf-8 -*-
"""
Useful functions for manipulating the action list and project structure.
"""

from copy import deepcopy
from pathlib import PurePath

from ..exceptions import ActionNotFound
from ..log import logger
from ..structure import FileOp, define_structure
from ..utils import get_id

logger = logger  # Sphinx workaround to force documenting imported members
"""Logger wrapper, that provides methods like :obj:`~.ReportLogger.report`.
See :class:`~.ReportLogger`.
"""

NO_OVERWRITE = FileOp.NO_OVERWRITE
"""Do not overwrite an existing file during update
(still created if not exists)
"""

NO_CREATE = FileOp.NO_CREATE
"""Do not create the file during an update"""


# -------- Project Structure --------

def _id_func(x):
    """Identity function"""
    return x


def modify(struct, path, modifier=_id_func, update_rule=None):
    """Modify the contents of a file in the representation of the project tree.

    If the given path, does not exist the parent directories are automatically
    created.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`. See :obj:`~.merge`.

        path (os.PathLike): path-like string or object relative to the
            structure root. The following examples are equivalent::

                from pathlib import PurePath

                'docs/api/index.html'
                PurePath('docs', 'api', 'index.html')

            *Deprecated* - Alternatively, a list with the parts of the path can
            be provided, ordered from the structure root to the file itself.

        modifier (callable): function (or callable object) that receives the
            old content as argument and returns the new content.
            If no modifier is passed, the identity function will be used.
            Note that, if the file does not exist in ``struct``, ``None`` will
            be passed as argument. Example::

                modifier = lambda old: (old or '') + 'APPENDED CONTENT'!
                modifier = lambda old: 'PREPENDED CONTENT!' + (old or '')

        update_rule: see :class:`~.FileOp`, ``None`` by default.
            Note that, if no ``update_rule`` is passed, the previous one is
            kept.

    Returns:
        dict: updated project tree representation

    Note:
        Use an empty string as content to ensure a file is created empty
        (``None`` contents will not be created).

    Warning:
        *Deprecation Notice* - In the next major release, the usage of lists
        for the ``path`` argument will result in an error. Please use
        :obj:`pathlib.PurePath` instead.
    """
    # Retrieve a list of parts from a path-like object
    if isinstance(path, (list, tuple)):
        path_parts = path
    else:
        # TODO: Remove conditional for v4 (always do the following)
        path_parts = PurePath(path).parts

    # Walk the entire path, creating parents if necessary.
    root = deepcopy(struct)
    last_parent = root
    name = path_parts[-1]
    for parent in path_parts[:-1]:
        last_parent = last_parent.setdefault(parent, {})

    # Get the old value if existent.
    old_value = last_parent.get(name, (None, None))
    if not isinstance(old_value, (list, tuple)):
        old_value = (old_value, None)

    # Update the value.
    new_value = (modifier(old_value[0]), update_rule)
    last_parent[name] = _merge_file_leaf(old_value, new_value)

    return root


def ensure(struct, path, content=None, update_rule=None):
    """Ensure a file exists in the representation of the project tree
    with the provided content.
    All the parent directories are automatically created.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`. See :obj:`~.merge`.

        path (os.PathLike): path-like string or object relative to the
            structure root. The following examples are equivalent::

                from pathlib import PurePath

                'docs/api/index.html'
                PurePath('docs', 'api', 'index.html')

            *Deprecated* - Alternatively, a list with the parts of the path can
            be provided, ordered from the structure root to the file itself.

        content (str): file text contents, ``None`` by default.
            The old content is preserved if ``None``.

        update_rule: see :class:`~.FileOp`, ``None`` by default

    Returns:
        dict: updated project tree representation

    Note:
        Use an empty string as content to ensure a file is created empty.

    Warning:
        *Deprecation Notice* - In the next major release, the usage of lists
        for the ``path`` argument will result in an error. Please use
        :obj:`pathlib.PurePath` instead.
    """
    modifier = _id_func if content is None else (lambda _: content)
    return modify(struct, path, modifier, update_rule)


def reject(struct, path):
    """Remove a file from the project tree representation if existent.

    Args:
        struct (dict): project representation as (possibly) nested
            :obj:`dict`. See :obj:`~.merge`.

        path (os.PathLike): path-like string or object relative to the
            structure root. The following examples are equivalent::

                from pathlib import PurePath

                'docs/api/index.html'
                PurePath('docs', 'api', 'index.html')

            *Deprecated* - Alternatively, a list with the parts of the path can
            be provided, ordered from the structure root to the file itself.

    Returns:
        dict: modified project tree representation

    Warning:
        *Deprecation Notice* - In the next major release, the usage of lists
        for the ``path`` argument will result in an error. Please use
        :obj:`pathlib.PurePath` instead.
    """
    # Retrieve a list of parts from a path-like object
    if isinstance(path, (list, tuple)):
        path_parts = path
    else:
        # TODO: Remove conditional for v4 (always do the following)
        path_parts = PurePath(path).parts

    # Walk the entire path, creating parents if necessary.
    root = deepcopy(struct)
    last_parent = root
    name = path_parts[-1]
    for parent in path_parts[:-1]:
        if parent not in last_parent:
            return root  # one ancestor already does not exist, do nothing
        last_parent = last_parent[parent]

    if name in last_parent:
        del last_parent[name]

    return root


def merge(old, new):
    """Merge two dict representations for the directory structure.

    Basically a deep dictionary merge, except from the leaf update method.

    Args:
        old (dict): directory descriptor that takes low precedence
                    during the merge
        new (dict): directory descriptor that takes high precedence
                    during the merge

    The directory tree is represented as a (possibly nested) dictionary.
    The keys indicate the path where a file will be generated, while the
    value indicates the content.  Additionally, tuple values are allowed in
    order to specify the rule that will be followed during an ``update``
    operation (see :class:`~.FileOp`).  In this case, the first element is
    the file content and the second element is the update rule. For
    example, the dictionary::

        {'project': {
            'namespace': {
                'module.py': ('print("Hello World!")',
                              helpers.NO_OVERWRITE)}}

    represents a ``project/namespace/module.py`` file with content
    ``print("Hello World!")``, that will be created only if not
    present.

    Returns:
        dict: resulting merged directory representation

    Note:
        Use an empty string as content to ensure a file is created empty.
        (``None`` contents will not be created).
    """
    return _inplace_merge(deepcopy(old), new)


def _inplace_merge(old, new):
    """Similar to :obj:`~.merge` but modifies the first dict."""

    for key, value in new.items():
        old_value = old.get(key, None)
        new_is_dict = isinstance(value, dict)
        old_is_dict = isinstance(old_value, dict)
        if new_is_dict and old_is_dict:
            old[key] = _inplace_merge(old_value, value)
        elif old_value is not None and not new_is_dict and not old_is_dict:
            # both are defined and final leaves
            old[key] = _merge_file_leaf(old_value, value)
        else:
            old[key] = deepcopy(value)

    return old


def _merge_file_leaf(old_value, new_value):
    """Merge leaf values for the directory tree representation.

    The leaf value is expected to be a tuple ``(content, update_rule)``.
    When a string is passed, it is assumed to be the content and
    ``None`` is used for the update rule.

    Args:
        old_value (tuple or str): descriptor for the file that takes low
                                  precedence during the merge
        new_value (tuple or str): descriptor for the file that takes high
                                  precedence during the merge

    Note:
        ``None`` contents are ignored, use and empty string to force empty
        contents.

    Returns:
        tuple or str: resulting value for the merged leaf
    """
    if not isinstance(old_value, (list, tuple)):
        old_value = (old_value, None)
    if not isinstance(new_value, (list, tuple)):
        new_value = (new_value, None)

    content = new_value[0] if new_value[0] is not None else old_value[0]
    rule = new_value[1] if new_value[1] is not None else old_value[1]

    if rule is None:
        return content

    return (content, rule)


# -------- Action List --------

def register(actions, action, before=None, after=None):
    """Register a new action to be performed during scaffold.

    Args:
        actions (list): previous action list.
        action (callable): function with two arguments: the first one is a
            (nested) dict representing the file structure of the project
            and the second is a dict with scaffold options.
            This function **MUST** return a tuple with two elements similar
            to its arguments. Example::

                def do_nothing(struct, opts):
                    return (struct, opts)

        **kwargs (dict): keyword arguments make it possible to choose a
            specific order when executing actions: when ``before`` or
            ``after`` keywords are provided, the argument value is used as
            a reference position for the new action. Example::

                helpers.register(actions, do_nothing,
                                 after='create_structure')
                    # Look for the first action with a name
                    # `create_structure` and inserts `do_nothing` after it.
                    # If more than one registered action is named
                    # `create_structure`, the first one is selected.

                helpers.register(
                    actions, do_nothing,
                    before='pyscaffold.structure:create_structure')
                    # Similar to the previous example, but the probability
                    # of name conflict is decreased by including the module
                    # name.

            When no keyword argument is provided, the default execution
            order specifies that the action will be performed after the
            project structure is defined, but before it is written to the
            disk. Example::


                helpers.register(actions, do_nothing)
                    # The action will take place after
                    # `pyscaffold.structure:define_structure`

    Returns:
        list: modified action list.
    """
    reference = before or after or get_id(define_structure)
    position = _find(actions, reference)
    if not before:
        position += 1

    clone = actions[:]
    clone.insert(position, action)

    return clone


def unregister(actions, reference):
    """Prevent a specific action to be executed during scaffold.

    Args:
        actions (list): previous action list.
        reference (str): action identifier. Similarly to the keyword
            arguments of :obj:`~.register` it can assume two formats:

                - the name of the function alone,
                - the name of the module followed by ``:`` and the name
                  of the function

    Returns:
        list: modified action list.
    """
    position = _find(actions, reference)
    return actions[:position] + actions[position+1:]


def _find(actions, name):
    """Find index of name in actions"""
    if ':' in name:
        names = [get_id(action) for action in actions]
    else:
        names = [action.__name__ for action in actions]

    try:
        return names.index(name)
    except ValueError:
        raise ActionNotFound(name)



# -*- coding: utf-8 -*-
"""
Extension that adjust project file tree to include a namespace package.

This extension adds a **namespace** option to
:obj:`~pyscaffold.api.create_project` and provides correct values for the
options **root_pkg** and **namespace_pkg** to the following functions in the
action list.
"""

import argparse
import os
from os.path import isdir
from os.path import join as join_path

from .. import templates, utils
from ..api import Extension, helpers
from ..log import logger


class Namespace(Extension):
    """Add a namespace (container package) to the generated package."""

    def augment_cli(self, parser):
        """Add an option to parser that enables the namespace extension.

        Args:
            parser (argparse.ArgumentParser): CLI parser object
        """
        parser.add_argument(
            self.flag,
            dest=self.name,
            default=None,
            action=create_namespace_parser(self),
            metavar="NS1[.NS2]",
            help="put your project inside a namespace package")

    def activate(self, actions):
        """Register an action responsible for adding namespace to the package.

        Args:
            actions (list): list of actions to perform

        Returns:
            list: updated list of actions
        """
        actions = helpers.register(actions, enforce_namespace_options,
                                   after='get_default_options')

        actions = helpers.register(actions, add_namespace,
                                   before='apply_update_rules')

        return helpers.register(actions, move_old_package,
                                after='create_structure')


def create_namespace_parser(obj_ref):
    """Create a namespace parser.

    Args:
        obj_ref (Extension): object reference to the actual extension

    Returns:
        NamespaceParser: parser for namespace cli argument
    """
    class NamespaceParser(argparse.Action):
        """Consumes the values provided, but also appends the extension
           function to the extensions list.
        """
        def __call__(self, parser, namespace, values, option_string=None):
            namespace.extensions.append(obj_ref)

            # Now the extra parameters can be stored
            setattr(namespace, self.dest, values)

            # save the namespace cli argument for later
            obj_ref.args = values

    return NamespaceParser


def enforce_namespace_options(struct, opts):
    """Make sure options reflect the namespace usage."""
    opts.setdefault('namespace', None)

    if opts['namespace']:
        opts['ns_list'] = utils.prepare_namespace(opts['namespace'])
        opts['root_pkg'] = opts['ns_list'][0]
        opts['qual_pkg'] = ".".join([opts['ns_list'][-1], opts['package']])

    return struct, opts


def add_namespace(struct, opts):
    """Prepend the namespace to a given file structure

    Args:
        struct (dict): directory structure as dictionary of dictionaries
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            directory structure as dictionary of dictionaries and input options
    """
    if not opts['namespace']:
        return struct, opts

    namespace = opts['ns_list'][-1].split('.')
    base_struct = struct
    struct = base_struct[opts['project']]['src']
    pkg_struct = struct[opts['package']]
    del struct[opts['package']]
    for sub_package in namespace:
        struct[sub_package] = {'__init__.py': templates.namespace(opts)}
        struct = struct[sub_package]
    struct[opts['package']] = pkg_struct

    return base_struct, opts


def move_old_package(struct, opts):
    """Move old package that may be eventually created without namespace

    Args:
        struct (dict): directory structure as dictionary of dictionaries
        opts (dict): options of the project

    Returns:
        tuple(dict, dict):
            directory structure as dictionary of dictionaries and input options
    """
    old_path = join_path(opts['project'], 'src', opts['package'])
    namespace_path = opts['qual_pkg'].replace('.', os.sep)
    target = join_path(opts['project'], 'src', namespace_path)

    old_exists = opts['pretend'] or isdir(old_path)
    #  ^  When pretending, pretend also an old folder exists
    #     to show a worst case scenario log to the user...

    if old_exists and opts['qual_pkg'] != opts['package']:
        if not opts['pretend']:
            logger.warning(
                '\nA folder %r exists in the project directory, and it is '
                'likely to have been generated by a PyScaffold extension or '
                'manually by one of the current project authors.\n'
                'Moving it to %r, since a namespace option was passed.\n'
                'Please make sure to edit all the files that depend on this '
                'package to ensure the correct location.\n',
                opts['package'], namespace_path)

        utils.move(old_path, target=target,
                   log=True, pretend=opts['pretend'])

    return struct, opts



# -*- coding: utf-8 -*-
"""
    Setup file for autodiffcst.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 3.2.3.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
import sys

from pkg_resources import VersionConflict, require
import setuptools

try:
    require('setuptools>=38.3')
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)



if __name__ == "__main__":
# setuptools.setup(use_pyscaffold=True)

# 

    # with open("docs/documentation.md", "r", encoding="utf-8") as fh:
    #     long_description = fh.read()

    with open('requirements.txt') as f:
        required = f.read().splitlines()
    

    setuptools.setup(
        name="autodiffCST", # Replace with your own username
        version="0.0.4",
        author="Xiaohan Yang, Hanwen Zhang, Runting Yang, Max Li",
        author_email="xiaohan_yang@g.harvard.edu, hzhang1@g.harvard.edu, runting_yang@hsph.harvard.edu, manli@fas.harvard.edu",
        description="This package autodiffCST implements automatic differentiation. Users could perform forward mode, and use it for higher order differentiation.",
        # long_description=long_description,
        # long_description_content_type="text/markdown",
        url="https://github.com/auto-differentiaters-in-CST/cs107-FinalProject",
        # packages=setuptools.find_packages(),
        packages=setuptools.find_packages(where="src"),
        package_dir = {'': "src"},
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.6',
        setup_requires=['pytest-runner'],
        install_requires=['numpy','sympy>=1.0'],
        tests_require=['pytest','coverage','pytest-cov'],
        test_suite="tests",
        py_modules = ['AD_vec', 'AD', "admath"]
    
        )


# Use a simple (but explicit) path modification to resolve the package properly
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))

import pytest
import math
import numpy as np
# import src.autodiffcst.AD as AD
# from src.autodiffcst.trigmath import *
import autodiffcst.AD as AD
from autodiffcst.AD_vec import *
import autodiffcst.admath as admath

# def test_set_VAD():
#     vadtest = VAD([1,2,3])
#     ad1 = vadtest[0]
#     ad2 = vadtest[1]
#     ad3 = vadtest[2]
#     new_vad = set_VAD(np.array([ad1,ad2,ad3]))
#     assert new_vad == vadtest, "Error: returned object not good."
#     assert sin(vadtest) == set_VAD(np.array([sin(ad1),sin(ad2),sin(ad3)])),"Error: returned object not good."

def test_abs():
    x = VAD([-1,2,3])
    f = abs(x)
    print(f.der)
    print(f.der2)
    assert np.sum(f.val == np.array([[1],[2],[3]])) == 3, "Error: abs didn't apply properly on VAD."
    assert np.sum(f.der == np.array([[-1., 0., 0.],[0., 1., 0.],[0., 0., 1.]])) == 9, "Error: der1 for abs(VAD) is not correct."
    x = AD.AD(-1, tag=0)
    f = abs(x)
    assert f.val == 1, "Error: abs didn't apply properly on AD."
    assert f.der == -1, "Error: der1 for abs(AD) is not correct."
    assert abs(-5) == 5, "Error: abs didn't apply properly on numbers."
    x = AD.AD(0, tag=0)
    with pytest.raises(Exception):
        f = abs(x)
    with pytest.raises(TypeError):
        f = admath.abs('error')
    xv = AD.AD(-2,tag=0,order=3)
    f = admath.abs(xv)
    assert f.higherdiff(3) == 0, "Error: abs didn't apply on higher properly"
    [x] = VAD([-5],order=3)
    f = abs(x)
    print(f.der)
    print(f.der2)
    print(f.higher)

def test_chain_rule():
    x = AD.AD(2,order=5)
    newad = 5*x**3+2
    higherde = np.array([60,60,30,0,0])
    adres = admath.chain_rule(x, 42, 60, 60, higher_der=higherde)
    assert newad == adres, "Error: chain rule doesn't apply to AD object properly."
    assert np.allclose(newad.higher,adres.higher), "Error: chain rule doesn't carry higher derivatives properly."
    # x = AD.AD(2,order=5)
    # newad = 5*x**3+2
    # higherde = np.array([60,60,30,0,0])
    # adres = chain_rule(x, 40, 60, 60, higher_der=higherde)
    # assert newad == adres, "Error: chain rule doesn't apply to AD object properly."
    # assert newad.higher == adres.higher, "Error: chain rule doesn't carry higher derivatives properly."

def test_choose():
    assert admath.choose(6,0) == 1, "Error: choose function doesn't calculate correctly."
    assert admath.choose(10,1) == 10, "Error: choose function doesn't calculate correctly."
    assert admath.choose(9,9) == 1, "Error: choose function doesn't calculate correctly."
    assert admath.choose(3,2) == 3, "Error: choose function doesn't calculate correctly."
    with pytest.raises(Exception):
        admath.choose(1,-1)
    with pytest.raises(Exception):
        admath.choose(5,11)
    with pytest.raises(Exception):
        admath.choose(4.5,1.1)


def test_log():
    x = AD.AD(1, tag=0,order=5)
    f = log(x)
    assert f.val[0] == np.log(1), "Error: log didn't apply properly on AD."
    assert np.allclose(f.der,1), "Error: der1 for log(AD) is not correct."
    assert np.allclose(f.der2,-1), "Error: der2 for log(AD) is not correct."
    higherarr = np.array([1,-1,2,-6,24])
    assert np.allclose(f.higher,higherarr), "Error: higherder for log(AD) is not correct."
    y = VAD([1,2,3])
    g = log(y)
    assert np.allclose(g.val, np.log([1,2,3])), "Error: log didn't apply properly on VAD."
    dertest = np.zeros((3,3))
    dertest[0,0] = 1
    dertest[1,1] = 1/2
    dertest[2,2] = 1/3
    assert np.allclose(g.der,dertest), "Error: der1 for log(VAD) is not correct."
    der2test = np.zeros((3,3,3))
    der2test[0,0,0] = -1**-2
    der2test[1,1,1] = -2**-2
    der2test[2,2,2] = -3**-2
    assert np.allclose(g.der2,der2test), "Error: der2 for log(VAD) is not correct."
    assert log(5) == np.log(5), "Error: log didn't apply properly on number."
    with pytest.raises(TypeError):
        f = admath.log('error')

def test_fact_ad():
    assert admath.fact_ad(2,0) == 1, "Error: fact_ad calculation wrong."
    assert admath.fact_ad(2,2) == 2, "Error: fact_ad calculation wrong."
    assert admath.fact_ad(2,3) == 0, "Error: fact_ad calculation wrong."
    assert admath.fact_ad(3.5,2) == 3.5*2.5, "Error: fact_ad calculation wrong."

def test_exp():
    x = AD.AD(1, tag=0,order=5)
    f = exp(x)
    assert f.val[0] == np.exp(1), "Error: exp didn't apply properly on AD."
    assert np.allclose(f.der,np.exp(1)), "Error: der1 for exp(AD) is not correct."
    assert np.allclose(f.der2,np.exp(1)), "Error: der2 for exp(AD) is not correct."
    higherarr = np.array([np.exp(1),np.exp(1),np.exp(1),np.exp(1),np.exp(1)])
    assert np.allclose(f.higher,higherarr), "Error: higherder for exp(AD) is not correct."
    y = VAD([1,2,3])
    g = exp(y)
    assert np.allclose(g.val, np.exp([1,2,3])), "Error: exp didn't apply properly on VAD."
    dertest = np.zeros((3,3))
    dertest[0,0] = np.exp(1)
    dertest[1,1] = np.exp(2)
    dertest[2,2] = np.exp(3)
    assert np.allclose(g.der,dertest), "Error: der1 for exp(VAD) is not correct."
    der2test = np.zeros((3,3,3))
    der2test[0,0,0] = np.exp(1)
    der2test[1,1,1] = np.exp(2)
    der2test[2,2,2] = np.exp(3)
    assert np.allclose(g.der2,der2test), "Error: der2 for exp(VAD) is not correct."
    assert exp(5) == np.exp(5), "Error: exp didn't apply properly on number."
    with pytest.raises(TypeError):
        f = admath.exp('error')

def test_sqrt():
    x = AD.AD(2, tag=0,order=3)
    f = sqrt(x)
    assert f.val[0] == np.sqrt(2), "Error: sqrt didn't apply properly on AD."
    assert np.allclose(f.der,0.5*1/np.sqrt(2)), "Error: der1 for sqrt(AD) is not correct."
    assert np.allclose(f.der2,-0.25/2/np.sqrt(2)), "Error: der2 for sqrt(AD) is not correct."
    assert np.abs(f.higherdiff(3) - 3/32/np.sqrt(2))<1e-8, "Error: higherder for sqrt(AD) is not correct."
    y = VAD([1,2])
    g = sqrt(y)
    assert np.allclose(g.val, np.sqrt([1,2])), "Error: sqrt didn't apply properly on VAD."
    dertest = np.zeros((2,2))
    dertest[0,0] = 0.5
    dertest[1,1] = 0.5*1/np.sqrt(2)
    assert np.allclose(g.der,dertest), "Error: der1 for sqrt(VAD) is not correct."
    der2test = np.zeros((2,2,2))
    der2test[0,0,0] = -0.25
    der2test[1,1,1] = -0.25/2/np.sqrt(2)
    assert np.allclose(g.der2,der2test), "Error: der2 for sqrt(VAD) is not correct."
    assert sqrt(5) == np.sqrt(5), "Error: sqrt didn't apply properly on number."
    with pytest.raises(TypeError):
        f = admath.sqrt('error')

# trig test:

def test_sin_float():
    ad = 2
    assert sin(ad).__eq__(np.sin(2)), "Error: sin(x), false value."
    with pytest.raises(TypeError):
        f = admath.sin('error')

def test_sin_ad():
    [x,y] = VAD([1,2])
    assert sin(x+y).val[0] == np.sin(3), "Error: sin(x+y), false value."
    assert np.allclose(jacobian(sin(x+y)),np.array([np.cos(3),np.cos(3)])), "Error: sin(x+y), false derivative."
    f = sin(VAD([1,2]))
    assert np.allclose(f.val,np.array([np.sin(1),np.sin(2)])), "Error: sin(VAD[1,2]), false value."
    assert np.allclose(sin(x+y).der2,np.array([[-np.sin(3),-np.sin(3)],[-np.sin(3),-np.sin(3)]])), "Error: sin(x+y), false der2."
    der1 = np.zeros((2,2))
    der1[0,0] = np.cos(1)
    der1[1,1] = np.cos(2)
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = -np.sin(1)
    der2[1,1,1] = -np.sin(2)
    assert np.allclose(f.der,der1), "Error: sin(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: sin(VAD[1,2]), false der2."
    [z] = VAD([np.pi/6],order=5)
    cosv = np.sqrt(3)/2
    sinv = 1/2
    higherde = np.array([cosv,-sinv,-cosv,sinv,cosv])
    h = sin(z)
    assert np.allclose(h.higher,higherde), "Error: sin(pi/6), false higher."

def test_cos_float():
    ad = 2
    assert cos(ad).__eq__(np.cos(2)), "Error: cos(x), false value."
    with pytest.raises(TypeError):
        f = admath.cos('error')

def test_cos_ad():
    [x,y] = VAD([1,2])
    assert cos(x+y).val[0] == np.cos(3), "Error: cos(x+y), false value."
    assert np.allclose(jacobian(cos(x+y)),np.array([-np.sin(3),-np.sin(3)])), "Error: cos(x+y), false derivative."
    f = cos(VAD([1,2]))
    assert np.allclose(cos(x+y).der2,np.array([[-np.cos(3),-np.cos(3)],[-np.cos(3),-np.cos(3)]])), "Error: cos(x+y), false der2."
    assert np.allclose(f.val,np.array([np.cos(1),np.cos(2)])), "Error: cos(VAD[1,2]), false value."
    der1 = np.zeros((2,2))
    der1[0,0] = -np.sin(1)
    der1[1,1] = -np.sin(2)
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = -np.cos(1)
    der2[1,1,1] = -np.cos(2)
    assert np.allclose(f.der,der1), "Error: cos(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: cos(VAD[1,2]), false der2."
    [z] = VAD([np.pi/6],order=5)
    cosv = np.sqrt(3)/2
    sinv = 1/2
    higherde = np.array([-sinv,-cosv,sinv,cosv,-sinv])
    h = cos(z)
    assert np.allclose(h.higher,higherde), "Error: cos(pi/6), false higher."

def test_tan_float():
    ad = 2
    assert tan(ad).__eq__(np.tan(2)), "Error: tan(x), false value."
    with pytest.raises(TypeError):
        f = admath.tan('error')

def test_tan_ad():
    [x,y] = VAD([1,2])
    assert tan(x+y).val[0] == np.tan(3), "Error: tan(x+y), false value."
    assert np.allclose(jacobian(tan(x+y)),np.array([admath.sec(3)**2,admath.sec(3)**2])), "Error: tan(x+y), false derivative."
    f = tan(VAD([1,2]))
    assert np.allclose(f.val,np.array([np.tan(1),np.tan(2)])), "Error: tan(VAD[1,2]), false value."
    der1 = np.zeros((2,2))
    assert np.allclose(tan(x+y).der2,np.array([[2*tan(3)*admath.sec(3)**2,2*tan(3)*admath.sec(3)**2],[2*tan(3)*admath.sec(3)**2,2*tan(3)*admath.sec(3)**2]])), "Error: tan(x+y), false der2."
    der1[0,0] = admath.sec(1)**2
    der1[1,1] = admath.sec(2)**2
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = 2*tan(1)*admath.sec(1)**2
    der2[1,1,1] = 2*tan(2)*admath.sec(2)**2
    assert np.allclose(f.der,der1), "Error: tan(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: tan(VAD[1,2]), false der2."
    [z] = VAD([1],order=3)
    h = tan(z)
    calc = 2*admath.sec(1)**2*(2*tan(1)**2+admath.sec(1)**2)
    assert np.abs(h.higherdiff(3)-calc)<1e-8, "Error: tan(VAD[1,2]), false higher."

def test_sec_float():
    ad = 2
    assert abs(admath.sec(ad)-1/math.cos(2)) <= 1e-8, "Error: sec(x), false value."
    with pytest.raises(TypeError):
        f = admath.sec('error')

# hyperbolic trig
def test_sinh_float():
    ad = 2
    assert abs(sinh(ad)-math.sinh(2)) <= 1e-8, "Error: sinh(x), false value."
    with pytest.raises(TypeError):
        f = admath.sinh('error')

def test_sinh_ad():
    [x,y] = VAD([1,2])
    assert sinh(x+y).val[0] == np.sinh(3), "Error: sinh(x+y), false value."
    assert np.allclose(jacobian(sinh(x+y)),np.array([np.cosh(3),np.cosh(3)])), "Error: sinh(x+y), false derivative."
    f = sinh(VAD([1,2]))
    assert np.allclose(sinh(x+y).der2,np.array([[np.sinh(3),np.sinh(3)],[np.sinh(3),np.sinh(3)]])), "Error: cosh(x+y), false der2."
    assert np.allclose(f.val,np.array([np.sinh(1),np.sinh(2)])), "Error: cosh(VAD[1,2]), false value."
    der1 = np.zeros((2,2))
    der1[0,0] = np.cosh(1)
    der1[1,1] = np.cosh(2)
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = np.sinh(1)
    der2[1,1,1] = np.sinh(2)
    assert np.allclose(f.der,der1), "Error: sinh(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: sinh(VAD[1,2]), false der2."
    [z] = VAD([0],order=5)
    cosv = 1.0
    sinv = 0.0
    higherde = np.array([cosv,sinv,cosv,sinv,cosv])
    h = sinh(z)
    assert np.allclose(h.higher,higherde), "Error: sinh(0), false higher."

def test_cosh_float():
    ad = 2
    assert abs(cosh(ad)-math.cosh(2)) <= 1e-8, "Error: cosh(x), false value."
    with pytest.raises(TypeError):
        f = admath.cosh('error')

def test_cosh_ad():
    [x,y] = VAD([1,2])
    assert cosh(x+y).val[0] == np.cosh(3), "Error: cosh(x+y), false value."
    assert np.allclose(jacobian(cosh(x+y)),np.array([np.sinh(3),np.sinh(3)])), "Error: cosh(x+y), false derivative."
    f = cosh(VAD([1,2]))
    assert np.allclose(cosh(x+y).der2,np.array([[np.cosh(3),np.cosh(3)],[np.cosh(3),np.cosh(3)]])), "Error: cosh(x+y), false der2."
    assert np.allclose(f.val,np.array([np.cosh(1),np.cosh(2)])), "Error: cosh(VAD[1,2]), false value."
    der1 = np.zeros((2,2))
    der1[0,0] = np.sinh(1)
    der1[1,1] = np.sinh(2)
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = np.cosh(1)
    der2[1,1,1] = np.cosh(2)
    assert np.allclose(f.der,der1), "Error: cosh(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: cosh(VAD[1,2]), false der2."
    [z] = VAD([0],order=5)
    cosv = 1.0
    sinv = 0.0
    higherde = np.array([sinv,cosv,sinv,cosv,sinv])
    h = cosh(z)
    assert np.allclose(h.higher,higherde), "Error: cosh(0), false higher."

def test_tanh_float():
    ad = 2
    assert abs(tanh(ad)-np.tanh(2)) <= 1e-8, "Error: tanh(x), false value."
    with pytest.raises(TypeError):
        f = admath.tanh('error')

def test_tanh_ad():
    [x,y] = VAD([1,2])
    assert tanh(x+y).val[0] == np.tanh(3), "Error: tanh(x+y), false value."
    assert np.allclose(jacobian(tanh(x+y)),np.array([admath.sech(3)**2,admath.sech(3)**2])), "Error: tanh(x+y), false derivative."
    assert np.allclose(tanh(x+y).der2,np.array([[-2*tanh(3)*admath.sech(3)**2,-2*tanh(3)*admath.sech(3)**2],[-2*tanh(3)*admath.sech(3)**2,-2*tanh(3)*admath.sech(3)**2]])), "Error: tanh(x+y), false der2."
    f = tanh(VAD([1,2]))
    assert np.allclose(f.val,np.array([np.tanh(1),np.tanh(2)])), "Error: tanh(VAD[1,2]), false value."
    der1 = np.zeros((2,2))
    der1[0,0] = admath.sech(1)**2
    der1[1,1] = admath.sech(2)**2
    der2 = np.zeros((2,2,2))
    der2[0,0,0] = -2*tanh(1)*admath.sech(1)**2
    der2[1,1,1] = -2*tanh(2)*admath.sech(2)**2
    assert np.allclose(f.der,der1), "Error: tanh(VAD[1,2]), false der1."
    assert np.allclose(f.der2,der2), "Error: tanh(VAD[1,2]), false der2."
    [z] = VAD([1],order=3)
    h = tanh(z)
    calc = 4*tanh(1)**2*admath.sech(1)**2 - 2*admath.sech(1)**4
    assert np.abs(h.higherdiff(3)-calc)<1e-8, "Error: tanh(VAD[1,2]), false higher."

def test_sech_float():
    ad = 2
    assert abs(admath.sech(ad)-1/np.cosh(2)) <= 1e-8, "Error: sech(x), false value."
    with pytest.raises(TypeError):
        f = admath.sech('error')



# -*- coding: utf-8 -*-
"""
Command-Line-Interface of PyScaffold
"""

import argparse
import logging
import os.path
import sys

from pkg_resources import parse_version

from . import __version__ as pyscaffold_version
from . import api, info, shell, templates, utils
from .exceptions import NoPyScaffoldProject
from .log import ReportFormatter, configure_logger
from .utils import get_id


def add_default_args(parser):
    """Add the default options and arguments to the CLI parser.

    Args:
        parser (argparse.ArgumentParser): CLI parser object
    """

    parser.add_argument(
        dest="project",
        help="project name",
        metavar="PROJECT")
    parser.add_argument(
        "-p",
        "--package",
        dest="package",
        required=False,
        help="package name (default: project name)",
        metavar="NAME")
    parser.add_argument(
        "-d",
        "--description",
        dest="description",
        required=False,
        help="package description",
        metavar="TEXT")
    license_choices = templates.licenses.keys()
    parser.add_argument(
        "-l",
        "--license",
        dest="license",
        choices=license_choices,
        required=False,
        default="mit",
        help="package license like {choices} (default: {default})".format(
            choices=', '.join(license_choices), default="mit"),
        metavar="LICENSE")
    parser.add_argument(
        "-u",
        "--url",
        dest="url",
        required=False,
        help="package url",
        metavar="URL")
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="force overwriting an existing directory")
    parser.add_argument(
        "-U",
        "--update",
        dest="update",
        action="store_true",
        default=False,
        help="update an existing project by replacing the most important files"
             " like setup.py etc. Use additionally --force to "
             "replace all scaffold files.")
    parser.add_argument(
        '-V',
        '--version',
        action='version',
        version='PyScaffold {ver}'.format(ver=pyscaffold_version))
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.INFO,
        dest="log_level",
        help="show additional information about current actions")
    parser.add_argument(
        "-vv",
        "--very-verbose",
        action="store_const",
        const=logging.DEBUG,
        dest="log_level",
        help="show all available information about current actions")

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-P",
        "--pretend",
        dest="pretend",
        action="store_true",
        default=False,
        help="do not create project, but displays the log of all operations"
             " as if it had been created.")
    group.add_argument(
        "--list-actions",
        dest="command",
        action="store_const",
        const=list_actions,
        help="do not create project, but show a list of planned actions")


def parse_args(args):
    """Parse command line parameters respecting extensions

    Args:
        args ([str]): command line parameters as list of strings

    Returns:
        dict: command line parameters
    """
    from pkg_resources import iter_entry_points

    # create the argument parser
    parser = argparse.ArgumentParser(
        description="PyScaffold is a tool for easily putting up the scaffold "
                    "of a Python project.")
    parser.set_defaults(log_level=logging.WARNING,
                        extensions=[],
                        command=run_scaffold)
    add_default_args(parser)
    # load and instantiate extensions
    cli_extensions = [extension.load()(extension.name) for extension
                      in iter_entry_points('pyscaffold.cli')]
    # add a group for mutually exclusive external generators
    mutex_group = parser.add_mutually_exclusive_group()
    for extension in cli_extensions:
        if extension.mutually_exclusive:
            extension.augment_cli(mutex_group)
        else:
            extension.augment_cli(parser)

    # Parse options and transform argparse Namespace object into common dict
    opts = vars(parser.parse_args(args))
    return opts


def process_opts(opts):
    """Process and enrich command line arguments

    Args:
        opts (dict): dictionary of parameters

    Returns:
        dict: dictionary of parameters from command line arguments
    """
    # When pretending the user surely wants to see the output
    if opts['pretend']:
        opts['log_level'] = logging.INFO

    configure_logger(opts)

    # In case of an update read and parse setup.cfg
    if opts['update']:
        try:
            opts = info.project(opts)
        except Exception as e:
            raise NoPyScaffoldProject from e

    # Save cli params for later updating
    opts['cli_params'] = {'extensions': list(), 'args': dict()}
    for extension in opts['extensions']:
        opts['cli_params']['extensions'].append(extension.name)
        if extension.args is not None:
            opts['cli_params']['args'][extension.name] = extension.args

    # Strip (back)slash when added accidentally during update
    opts['project'] = opts['project'].rstrip(os.sep)

    # Remove options with None values
    opts = {k: v for k, v in opts.items() if v is not None}
    return opts


def run_scaffold(opts):
    """Actually scaffold the project, calling the python API

    Args:
        opts (dict): command line options as dictionary
    """
    api.create_project(opts)
    if opts['update'] and not opts['force']:
        note = "Update accomplished!\n" \
               "Please check if your setup.cfg still complies with:\n" \
               "https://pyscaffold.org/en/v{}/configuration.html"
        base_version = parse_version(pyscaffold_version).base_version
        print(note.format(base_version))


def list_actions(opts):
    """Do not create a project, just list actions considering extensions

    Args:
        opts (dict): command line options as dictionary
    """
    actions = api.discover_actions(opts.get('extensions', []))

    print('Planned Actions:')
    for action in actions:
        print(ReportFormatter.SPACING + get_id(action))


def main(args):
    """Main entry point for external applications

    Args:
        args ([str]): command line arguments
    """
    utils.check_setuptools_version()
    opts = parse_args(args)
    opts = process_opts(opts)
    opts['command'](opts)


@shell.shell_command_error2exit_decorator
@utils.exceptions2exit([RuntimeError])
def run():
    """Entry point for console script"""
    main(sys.argv[1:])


if __name__ == '__main__':
    main(sys.argv[1:])
